#include "command.h"
#include "render.h"
#include "arpa/inet.h"
#include "ui.h"
#include "vk_mem_alloc.h"
#include "vulkan/vk_enum_string_helper.h"
#include "vulkan/vulkan_core.h"

#include "ft2build.h"
#include FT_FREETYPE_H

VkResult render_thread(GLFWwindow* window, RenderContext* render_context) {
  FT_Library library;

  UIContextStorage ui;
  UILayerStorage layers[2];

  FontStorage font;

  VkBuffer transfer;
  VmaAllocation transfer_memory;
  void* mapped;

  VkResult result;

  VK_RESULT(create_ui_context(
        render_context->device,
        render_context->allocator,
        render_context->render_pass,
        10,
        render_context->swapchain_extent,
        render_context->window_scale,
        render_context->transfer_pool,
        render_context->transfer_queue,
        &ui));

  if(FT_Init_FreeType(&library) != FT_Err_Ok) {
    return VK_ERROR_UNKNOWN;
  }

  VK_RESULT(load_font(render_context->device, render_context->allocator, ui.font_infos, ui.font_samplers, ui.font_textures, 0, render_context->transfer_pool, render_context->transfer_queue, library, "test.ttf", 16, VK_TRUE, &font));

  VK_RESULT(create_layer(10, 100, 10, render_context->device, render_context->allocator, render_context->transfer_pool, render_context->transfer_queue, &layers[0]));
  VK_RESULT(create_layer(10, 100, 10, render_context->device, render_context->allocator, render_context->transfer_pool, render_context->transfer_queue, &layers[1]));

  VK_RESULT(create_transfer_buffer(render_context->allocator, sizeof(uint32_t) + 2*sizeof(UIRect), &transfer, &transfer_memory, &mapped));

  uint32_t* mapped_count = (uint32_t*)(mapped);
  mapped_count[0] = 1;
  UIRect* mapped_rect = (UIRect*)(mapped + sizeof(uint32_t));
  mapped_rect[0].pos[0] = 0.0;
  mapped_rect[0].pos[1] = 0.0;
  mapped_rect[0].size[0] = 200.0;
  mapped_rect[0].size[1] = 200.0;
  mapped_rect[0].color[0] = 1.0;
  mapped_rect[0].color[1] = 0.0;
  mapped_rect[0].color[2] = 0.0;
  mapped_rect[0].color[3] = 1.0;

  mapped_rect[1].pos[0] = 200.0;
  mapped_rect[1].pos[1] = 0.0;
  mapped_rect[1].size[0] = 200.0;
  mapped_rect[1].size[1] = 200.0;
  mapped_rect[1].color[0] = 0.0;
  mapped_rect[1].color[1] = 1.0;
  mapped_rect[1].color[2] = 0.0;
  mapped_rect[1].color[3] = 1.0;

  VkCommandBuffer command_buffer = command_begin_single(render_context->device, render_context->transfer_pool);
  command_copy_buffer(command_buffer, transfer, layers[0].layer, 0, offsetof(UILayer, draw_rects) + offsetof(DrawCommand, instance_count), sizeof(uint32_t));
  command_copy_buffer(command_buffer, transfer, layers[1].layer, 0, offsetof(UILayer, draw_rects) + offsetof(DrawCommand, instance_count), sizeof(uint32_t));
  command_copy_buffer(command_buffer, transfer, layers[0].rects, sizeof(uint32_t), 0, sizeof(UIRect));
  command_copy_buffer(command_buffer, transfer, layers[1].rects, sizeof(uint32_t) + sizeof(UIRect), 0, sizeof(UIRect));
  VK_RESULT(command_end_single(render_context->device, command_buffer, render_context->transfer_pool, render_context->transfer_queue));
  vkQueueWaitIdle(render_context->transfer_queue.handle);
  destroy_transfer_buffer(render_context->allocator, transfer, transfer_memory);

  while(glfwWindowShouldClose(window) == 0) {
    glfwPollEvents();

    result = draw_frame(render_context, &ui, layers, 2);
    if(result != VK_SUCCESS) {
      fprintf(stderr, "draw_frame error: %s\n", string_VkResult(result));
      return result;
    }
  }
  return 0;
}

int logic_thread() {
  return 0;
}

int network_thread() {
  return 0;
}

int main() {
  GLFWwindow* window = init_window();
  if(window == NULL) {
    return 1;
  }

  RenderContext render_context = {};
  if(init_vulkan(window, &render_context) != VK_SUCCESS) {
    return 2;
  }

  if(render_thread(window, &render_context) != VK_SUCCESS) {
    return 3;
  }

  return 0;
}
